from libc.stdint cimport int16_t
from libcpp cimport bool
from libcpp.string cimport string


cdef extern from "sink.h" namespace "memray::io":
    cdef cppclass Sink:
        pass

    cdef cppclass FileSink(Sink):
        FileSink(const string& file_name, bool overwrite, bool compress) except +IOError

    cdef cppclass SocketSink(Sink):
        SocketSink(string host, unsigned int port) except +IOError

    cdef cppclass NullSink(Sink):
        NullSink() except +IOError



from _memray.hooks cimport Allocator
from _memray.records cimport AggregatedAllocation
from _memray.records cimport Allocation
from _memray.records cimport optional_frame_id_t
from libc.stdint cimport uint64_t
from libcpp cimport bool
from libcpp.functional cimport function
from libcpp.unordered_map cimport unordered_map
from libcpp.utility cimport pair
from libcpp.vector cimport vector


cdef extern from "snapshot.h" namespace "memray::api":
    unsigned long NO_THREAD_INFO

    cdef struct HighWatermark:
        size_t index
        size_t peak_memory

    cdef cppclass HighWatermarkFinder:
        void processAllocation(const Allocation&) except+
        HighWatermark getHighWatermark()
        size_t getCurrentWatermark()

    cdef cppclass reduced_snapshot_map_t:
        pass

    cdef cppclass AbstractAggregator:
        void addAllocation(const Allocation&) except+
        reduced_snapshot_map_t getSnapshotAllocations(bool merge_threads) except+

    cdef cppclass TemporaryAllocationsAggregator(AbstractAggregator):
        TemporaryAllocationsAggregator(size_t max_items)

    cdef cppclass SnapshotAllocationAggregator(AbstractAggregator):
        pass

    cdef cppclass AggregatedCaptureReaggregator(AbstractAggregator):
        pass

    cdef cppclass LocationKey:
        size_t python_frame_id
        size_t native_frame_id
        unsigned long thread_id

    cdef cppclass index_thread_pair_hash:
        pass

    cdef cppclass HighWaterMarkLocationKey:
        unsigned long thread_id
        size_t python_frame_id
        size_t native_frame_id
        size_t native_segment_generation
        Allocator allocator

        bool operator==(const HighWaterMarkLocationKey& other)
        bool operator!=(const HighWaterMarkLocationKey& other)

    cdef cppclass AllocationLifetime:
        size_t allocatedBeforeSnapshot
        size_t deallocatedBeforeSnapshot
        HighWaterMarkLocationKey key
        size_t n_allocations
        size_t n_bytes

    cdef cppclass AllocationLifetimeAggregator:
        void addAllocation(const Allocation& allocation) except+
        void captureSnapshot()
        vector[AllocationLifetime] generateIndex() except+

    cdef cppclass HighWaterMarkAggregator:
        void addAllocation(const Allocation& allocation) except+
        void captureSnapshot() except+

        size_t getCurrentHeapSize()
        bool visitAllocations[T](const T& callback) except+
        vector[size_t] highWaterMarkBytesBySnapshot() except+
        vector[AllocationLifetime] generateIndex() except+

    cdef cppclass AllocationStatsAggregator:
        void addAllocation(const Allocation&, optional_frame_id_t python_frame_id) except+
        uint64_t totalAllocations()
        uint64_t totalBytesAllocated()
        uint64_t peakBytesAllocated()
        const unordered_map[size_t, uint64_t]& allocationCountBySize()
        const unordered_map[int, uint64_t]& allocationCountByAllocator()
        vector[pair[uint64_t, optional_frame_id_t]] topLocationsBySize(size_t num_largest) except+
        vector[pair[uint64_t, optional_frame_id_t]] topLocationsByCount(size_t num_largest) except+

    object Py_ListFromSnapshotAllocationRecords(const reduced_snapshot_map_t&) except+
    object Py_GetSnapshotAllocationRecords(const vector[Allocation]& all_records, size_t record_index, bool merge_threads) except+



from _memray.records cimport Allocation
from libcpp cimport bool


cdef extern from "hooks.h" namespace "memray::hooks":
    cdef cppclass Allocator:
        pass

    bool isDeallocator(const Allocator& allocator)



cdef extern from "<pthread.h>" nogil:
    ctypedef int pthread_t

    ctypedef struct pthread_attr_t:
        pass
    ctypedef struct pthread_mutexattr_t:
        pass
    ctypedef struct pthread_mutex_t:
       pass

    enum:
        PTHREAD_CANCEL_ENABLE
        PTHREAD_CANCEL_DISABLE

    int pthread_cancel(pthread_t thread)
    int pthread_setcancelstate(int state, int *oldstate)
    pthread_t pthread_self()
    int pthread_equal(pthread_t t1, pthread_t t2)
    int pthread_create(pthread_t *thread, pthread_attr_t *attr,
                       void *(*start_routine) (void *), void *arg)
    int pthread_join(pthread_t thread, void **retval)
    int pthread_kill(pthread_t thread, int sig)



from _memray.hooks cimport Allocator
from libc.stdint cimport uintptr_t
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector


cdef extern from "records.h" namespace "memray::tracking_api":
   ctypedef unsigned long thread_id_t
   ctypedef size_t frame_id_t

   struct Frame:
       string function_name
       string filename
       int lineno

   struct TrackerStats:
       size_t n_allocations
       size_t n_frames
       long long start_time
       long long end_time

   cdef enum FileFormat:
       ALL_ALLOCATIONS 'memray::tracking_api::FileFormat::ALL_ALLOCATIONS'
       AGGREGATED_ALLOCATIONS 'memray::tracking_api::FileFormat::AGGREGATED_ALLOCATIONS'

   struct HeaderRecord:
       int version
       bool native_traces
       FileFormat file_format
       TrackerStats stats
       string command_line
       int pid
       size_t main_tid
       size_t skipped_frames_on_main_tid
       int python_allocator
       bool trace_python_allocators

   cdef cppclass Allocation:
       thread_id_t tid
       uintptr_t address
       size_t size
       Allocator allocator
       frame_id_t native_frame_id
       size_t frame_index
       size_t native_segment_generation
       size_t n_allocations

       object toPythonObject()

   cdef cppclass AggregatedAllocation:
       thread_id_t tid
       Allocator allocator
       frame_id_t native_frame_id
       size_t frame_index
       size_t native_segment_generation

       size_t n_allocations_in_high_water_mark
       size_t n_allocations_leaked
       size_t bytes_in_high_water_mark
       size_t bytes_leaked

       Allocation contributionToHighWaterMark()
       Allocation contributionToLeaks()

   struct MemoryRecord:
       unsigned long int ms_since_epoch
       size_t rss

   struct MemorySnapshot:
       unsigned long int ms_since_epoch
       size_t rss
       size_t heap


cdef extern from "<optional>":
   # Cython doesn't have libcpp.optional yet, so just declare this opaquely.
   cdef cppclass optional_frame_id_t "std::optional<memray::tracking_api::frame_id_t>":
       pass



from ..reporters.table import TableReporter
from .common import HighWatermarkCommand


class TableCommand(HighWatermarkCommand):
    """Generate an HTML table with all records in the peak memory usage"""

    def __init__(self) -> None:
        super().__init__(
            reporter_factory=TableReporter.from_snapshot,
            reporter_name="table",
        )



from libcpp cimport bool
from libcpp.string cimport string


cdef extern from "source.h" namespace "memray::io":
    cdef cppclass Source:
        pass

    cdef cppclass FileSource(Source):
        FileSource(const string& file_name) except+ IOError

    cdef cppclass SocketSource(Source):
        SocketSource(int port) except+ IOError



from _memray.record_writer cimport RecordWriter
from libc.stdint cimport uint64_t
from libcpp cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string


cdef extern from "tracking_api.h" namespace "memray::tracking_api":
    void set_up_pthread_fork_handlers() except+
    void install_trace_function() except*

    cdef cppclass Tracker:
        @staticmethod
        object createTracker(
            unique_ptr[RecordWriter] record_writer,
            bool native_traces,
            unsigned int memory_interval,
            bool follow_fork,
            bool trace_pymalloc,
        ) except+

        @staticmethod
        object destroyTracker() except +

        @staticmethod
        Tracker* getTracker()

        @staticmethod
        void forgetPythonStack() except+

        @staticmethod
        void beginTrackingGreenlets() except+

        @staticmethod
        void handleGreenletSwitch(object, object) except+

        @staticmethod
        void registerThreadNameById(uint64_t, const char*) except+

        @staticmethod
        void childFork() noexcept nogil



from _memray.records cimport AggregatedAllocation
from _memray.records cimport Allocation
from _memray.records cimport HeaderRecord
from _memray.records cimport MemoryRecord
from _memray.records cimport MemorySnapshot
from _memray.records cimport optional_frame_id_t
from _memray.source cimport Source
from libcpp cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string
from libcpp.vector cimport vector


cdef extern from "record_reader.h" namespace "memray::api":
    cdef enum RecordResult 'memray::api::RecordReader::RecordResult':
        RecordResultAllocationRecord 'memray::api::RecordReader::RecordResult::ALLOCATION_RECORD'
        RecordResultAggregatedAllocationRecord 'memray::api::RecordReader::RecordResult::AGGREGATED_ALLOCATION_RECORD'
        RecordResultMemoryRecord 'memray::api::RecordReader::RecordResult::MEMORY_RECORD'
        RecordResultMemorySnapshot 'memray::api::RecordReader::RecordResult::MEMORY_SNAPSHOT'
        RecordResultError 'memray::api::RecordReader::RecordResult::ERROR'
        RecordResultEndOfFile 'memray::api::RecordReader::RecordResult::END_OF_FILE'

    cdef cppclass RecordReader:
        RecordReader(unique_ptr[Source]) except+
        RecordReader(unique_ptr[Source], bool track_stacks) except+
        void close()
        bool isOpen() const
        RecordResult nextRecord() except+
        object Py_GetStackFrame(unsigned int frame_id) except+
        object Py_GetStackFrame(unsigned int frame_id, size_t max_stacks) except+
        object Py_GetStackFrameAndEntryInfo(
            unsigned int frame_id, vector[unsigned char]* is_entry_frame
        ) except+
        object Py_GetStackFrameAndEntryInfo(
            unsigned int frame_id, vector[unsigned char]* is_entry_frame, size_t max_stacks
        ) except+
        object Py_GetNativeStackFrame(int frame_id, size_t generation) except+
        object Py_GetNativeStackFrame(int frame_id, size_t generation, size_t max_stacks) except+
        optional_frame_id_t getLatestPythonFrameId(const Allocation&) except+
        object Py_GetFrame(optional_frame_id_t frame) except+
        HeaderRecord getHeader()
        size_t getMainThreadTid()
        size_t getSkippedFramesOnMainThread()
        object dumpAllRecords() except+
        string getThreadName(long int tid) except+
        Allocation getLatestAllocation()
        MemoryRecord getLatestMemoryRecord()
        AggregatedAllocation getLatestAggregatedAllocation()
        MemorySnapshot getLatestMemorySnapshot()



import distutils.command.build
import distutils.log
import os
import pathlib
import subprocess
import sys
import tempfile
from sys import platform
from sys import version_info

import pkgconfig
from Cython.Build import cythonize
from setuptools import Extension
from setuptools import find_packages
from setuptools import setup
from setuptools.command.build_ext import build_ext as build_ext_orig

IS_MAC = sys.platform == "darwin"
IS_LINUX = "linux" in sys.platform

LIBBACKTRACE_LOCATION = (
    pathlib.Path(__file__).parent / "src" / "vendor" / "libbacktrace"
).resolve()

LIBBACKTRACE_INCLUDEDIRS = LIBBACKTRACE_LOCATION / "install" / "include"
LIBBACKTRACE_LIBDIR = LIBBACKTRACE_LOCATION / "install" / "lib"

ASSETS_LOCATION = (
    pathlib.Path(__file__).parent
    / "src"
    / "memray"
    / "reporters"
    / "templates"
    / "assets"
)


class BuildMemray(build_ext_orig):
    def run(self):
        self.build_js_files()
        self.build_libbacktrace()
        super().run()

    def announce_and_run(self, command, **kwargs):
        self.announce(
            "executing command: `{}`".format(" ".join(command)),
            level=distutils.log.INFO,
        )
        subprocess.run(command, check=True, **kwargs)

    def build_libbacktrace(self):
        archive_location = LIBBACKTRACE_LIBDIR / "libbacktrace.a"

        if archive_location.exists():
            return

        if not LIBBACKTRACE_LOCATION.exists():
            self.announce_and_run(
                [f"{LIBBACKTRACE_LOCATION.parent / 'regenerate_libbacktrace.sh'}"],
                cwd=LIBBACKTRACE_LOCATION.parent,
            )

        configure_cmd = [
            f"{LIBBACKTRACE_LOCATION}/configure",
            "--with-pic",
            "--prefix",
            f"{LIBBACKTRACE_LOCATION}/install",
            "--includedir",
            f"{LIBBACKTRACE_LOCATION}/install/include/libbacktrace",
        ]
        libbacktrace_target = os.getenv("MEMRAY_LIBBACKTRACE_TARGET")
        if libbacktrace_target is not None:
            configure_cmd.extend(["--host", libbacktrace_target])

        with tempfile.TemporaryDirectory() as tmpdirname:
            self.announce_and_run(
                configure_cmd,
                cwd=tmpdirname,
            )
            self.announce_and_run(["make", "-j"], cwd=tmpdirname)
            self.announce_and_run(["make", "install"], cwd=tmpdirname)

    def build_js_files(self):
        if any(ASSETS_LOCATION.glob("*.js")):
            return

        self.announce_and_run(["npm", "install"])
        self.announce_and_run(["npm", "run-script", "build"])


install_requires = [
    "jinja2 >= 2.9",
    "typing_extensions; python_version < '3.8.0'",
    "rich >= 11.2.0",
    "textual >= 0.41.0",
]
docs_requires = [
    "IPython",
    "bump2version",
    "sphinx",
    "furo",
    "sphinx-argparse",
    "towncrier",
]

lint_requires = [
    "black",
    "flake8",
    "isort",
    "mypy",
    "check-manifest",
]

test_requires = [
    "Cython",
    "greenlet; python_version < '3.14'",
    "pytest",
    "pytest-cov",
    "ipython",
    "setuptools; python_version >= '3.12'",
    "pytest-textual-snapshot",
    "textual >= 0.43, != 0.65.2, != 0.66",
    "packaging",
]

benchmark_requires = [
    "asv",
]


TEST_BUILD = False
if "--test-build" in sys.argv:
    TEST_BUILD = True
    sys.argv.remove("--test-build")


if os.getenv("CYTHON_TEST_MACROS", None) is not None:
    TEST_BUILD = True

MINIMIZE_INLINING = os.getenv("MEMRAY_MINIMIZE_INLINING", "") != ""

COMPILER_DIRECTIVES = {
    "language_level": 3,
    "embedsignature": True,
    "boundscheck": False,
    "wraparound": False,
    "cdivision": True,
    "profile": False,
    "linetrace": False,
    "c_string_type": "unicode",
    "c_string_encoding": "utf8",
}
EXTRA_COMPILE_ARGS = []
EXTRA_LINK_ARGS = []
UNDEF_MACROS = []

if MINIMIZE_INLINING:
    EXTRA_COMPILE_ARGS.append("-Og")
else:
    EXTRA_COMPILE_ARGS.append("-flto")
    EXTRA_LINK_ARGS.append("-flto")

# For Python 3.9+, hide all of our symbols except the module init function. For
# Python 3.8 and earlier this isn't as easy, because PyMODINIT_FUNC doesn't
# include __attribute__((visibility ("default"))), and Cython doesn't give us
# a way to add the attribute. So, skip this optimization on 3.8 and earlier.
if sys.version_info[:2] >= (3, 9):
    EXTRA_COMPILE_ARGS.append("-fvisibility=hidden")

if TEST_BUILD:
    COMPILER_DIRECTIVES = {
        "language_level": 3,
        "boundscheck": True,
        "embedsignature": True,
        "wraparound": True,
        "cdivision": False,
        "profile": False,
        "linetrace": False,
        "overflowcheck": True,
        "infer_types": True,
        "c_string_type": "unicode",
        "c_string_encoding": "utf8",
    }
    EXTRA_COMPILE_ARGS = []
    UNDEF_MACROS = ["NDEBUG"]
    if IS_LINUX:
        EXTRA_COMPILE_ARGS.extend(["-D_GLIBCXX_DEBUG", "-D_LIBCPP_DEBUG"])

DEFINE_MACROS = []

# Ensure that we have a 64-bit off_t in all translation units.
DEFINE_MACROS.append(("_FILE_OFFSET_BITS", "64"))

# memray uses thread local storage (TLS) variables. As memray is compiled
# into a Python extension, is a shared object. TLS variables in shared objects
# use the most conservative and slow TLS model available by default:
# global-dynamic. This TLS model generates function calls (__tls_get_addr) to
# obtain the address of the TLS storage block, which is quite slow.  To
# circuvent the slowdown, memray uses by default a less restrictive model:
# initial-exec. This model is very fast but uses the limited TLS storage of the
# executable. This means that is possible that dlopen will refuse to load the
# shared object of the extension if there is not enough space. glibc reserves
# 1152 bytes for oportunustic usage for shared libraries with initial-exec, so
# this model will not present problems as long as the application uses glibc. In
# case these assumptions are wrong, memray can revert to use the most
# conservative model by setting the NO_MEMRAY_FAST_TLS environment variable.

MEMRAY_FAST_TLS = True
if os.getenv("NO_MEMRAY_FAST_TLS", None) is not None:
    MEMRAY_FAST_TLS = False

if MEMRAY_FAST_TLS:
    DEFINE_MACROS.append(("USE_MEMRAY_TLS_MODEL", "1"))

BINARY_FORMATS = {"darwin": "macho", "linux": "elf"}
BINARY_FORMAT = BINARY_FORMATS.get(sys.platform, "elf")

library_flags = {"libraries": ["lz4"]}
if IS_LINUX:
    library_flags["libraries"].append("unwind")
    library_flags["libraries"].append("debuginfod")

try:
    library_flags = pkgconfig.parse(
        " ".join(f"lib{libname}" for libname in library_flags["libraries"])
    )
except EnvironmentError as e:
    print("pkg-config not found.", e)
    print("Falling back to static flags.")
except pkgconfig.PackageNotFoundError as e:
    print("Package Not Found", e)
    print("Falling back to static flags.")

MEMRAY_EXTENSION = Extension(
    name="memray._memray",
    sources=[
        "src/memray/_memray.pyx",
        "src/memray/_memray/compat.cpp",
        "src/memray/_memray/hooks.cpp",
        "src/memray/_memray/tracking_api.cpp",
        f"src/memray/_memray/{BINARY_FORMAT}_shenanigans.cpp",
        "src/memray/_memray/logging.cpp",
        "src/memray/_memray/python_helpers.cpp",
        "src/memray/_memray/source.cpp",
        "src/memray/_memray/sink.cpp",
        "src/memray/_memray/records.cpp",
        "src/memray/_memray/record_reader.cpp",
        "src/memray/_memray/record_writer.cpp",
        "src/memray/_memray/snapshot.cpp",
        "src/memray/_memray/socket_reader_thread.cpp",
        "src/memray/_memray/native_resolver.cpp",
    ],
    language="c++",
    extra_compile_args=["-std=c++17", "-Wall", *EXTRA_COMPILE_ARGS],
    extra_objects=[str(LIBBACKTRACE_LIBDIR / "libbacktrace.a")],
    extra_link_args=["-std=c++17", *EXTRA_LINK_ARGS],
    define_macros=DEFINE_MACROS,
    undef_macros=UNDEF_MACROS,
    **library_flags,
)

MEMRAY_EXTENSION.include_dirs[:0] = ["src", str(LIBBACKTRACE_INCLUDEDIRS)]
MEMRAY_EXTENSION.libraries.append("dl")


MEMRAY_TEST_EXTENSION = Extension(
    name="memray._test_utils",
    sources=[
        "src/memray/_memray_test_utils.pyx",
    ],
    language="c++",
    extra_compile_args=["-std=c++17", "-Wall", *EXTRA_COMPILE_ARGS],
    extra_link_args=["-std=c++17", *EXTRA_LINK_ARGS],
    define_macros=DEFINE_MACROS,
    undef_macros=UNDEF_MACROS,
)

MEMRAY_INJECT_EXTENSION = Extension(
    name="memray._inject",
    sources=[
        "src/memray/_memray/inject.cpp",
    ],
    language="c++",
    extra_compile_args=["-std=c++17", "-Wall", *EXTRA_COMPILE_ARGS],
    extra_link_args=["-std=c++17", *EXTRA_LINK_ARGS],
    define_macros=DEFINE_MACROS,
    undef_macros=UNDEF_MACROS,
    py_limited_api=True,
)


if not (IS_LINUX or IS_MAC):
    raise RuntimeError(f"memray does not support this platform ({platform})")

about = {}
with open("src/memray/_version.py") as fp:
    exec(fp.read(), about)


HERE = pathlib.Path(__file__).parent.resolve()
LONG_DESCRIPTION = (HERE / "README.md").read_text(encoding="utf-8")

setup(
    name="memray",
    version=about["__version__"],
    python_requires=">=3.7.0",
    description="A memory profiler for Python applications",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/bloomberg/memray",
    author="Pablo Galindo Salgado",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Debuggers",
    ],
    license="Apache 2.0",
    package_dir={"": "src"},
    packages=find_packages(where="src", exclude="memray/_memray/"),
    ext_modules=cythonize(
        [MEMRAY_EXTENSION, MEMRAY_TEST_EXTENSION, MEMRAY_INJECT_EXTENSION],
        include_path=["src/memray"],
        compiler_directives=COMPILER_DIRECTIVES,
    ),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        "test": test_requires,
        "docs": docs_requires,
        "lint": lint_requires,
        "benchmark": benchmark_requires,
        "dev": test_requires + lint_requires + docs_requires + benchmark_requires,
    },
    entry_points={
        "console_scripts": [
            f"memray{version_info.major}.{version_info.minor}=memray.__main__:main",
            "memray=memray.__main__:main",
        ],
    },
    cmdclass={
        "build_ext": BuildMemray,
    },
)



from _memray.record_reader cimport RecordReader
from libcpp cimport bool
from libcpp cimport int
from libcpp.memory cimport shared_ptr


cdef extern from "socket_reader_thread.h" namespace "memray::socket_thread":
    cdef cppclass BackgroundSocketReader:
        BackgroundSocketReader(shared_ptr[RecordReader]) except+

        void start() except+
        bool is_active()
        object Py_GetSnapshotAllocationRecords(bool merge_threads)



from _memray.records cimport FileFormat
from _memray.sink cimport Sink
from libcpp cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string


cdef extern from "record_writer.h" namespace "memray::api":
    cdef cppclass RecordWriter:
        pass
    cdef unique_ptr[RecordWriter] createRecordWriter(
        unique_ptr[Sink],
        string command_line,
        bool native_trace,
        FileFormat file_format,
        bool trace_python_allocators,
    ) except+



cdef extern from "alloc.h" nogil:
    void *calloc (size_t count, size_t eltsize)
    void free (void *ptr)
    void *malloc (size_t size)
    int posix_memalign(void** memptr, size_t alignment, size_t size)
    void* aligned_alloc(size_t alignment, size_t size)
    void *realloc (void *ptr, size_t newsize)
    void* valloc(size_t size)
    void* memalign(size_t alignment, size_t size)
    void* pvalloc(size_t size)

cdef extern from "Python.h":
    void* PyMem_RawMalloc(size_t n) nogil
    void* PyMem_RawCalloc(size_t nelem, size_t elsize) nogil
    void* PyMem_RawRealloc(void *p, size_t n) nogil
    void PyMem_RawFree(void *p) nogil

    void* PyMem_Malloc(size_t n)
    void* PyMem_Calloc(size_t nelem, size_t elsize)
    void* PyMem_Realloc(void *p, size_t n)
    void PyMem_Free(void *p)

    void* PyObject_Malloc(size_t size)
    void* PyObject_Calloc(size_t nelem, size_t elsize)
    void* PyObject_Realloc(void *ptr, size_t new_size)
    void PyObject_Free(void *ptr)
